from maplib import Model

gr1 = """
<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "A" .
"""

gr2 = """
<http://example.net/ns#myObject> <http://example.net/ns#hasValue> "B" . 
"""


def test_basic_named_graph():
    ng2 = "urn:graph:gr2"
    m = Model()
    m.reads(gr1, format="ntriples")
    m.reads(gr2, format="ntriples", graph=ng2)

    df = m.query("""
    SELECT * WHERE {
        GRAPH <urn:graph:gr2> {
            ?a ?b ?c .
        }
    }    
    """)
    assert df.height == 1
    assert df.get_column("c")[0] == "B"

def test_basic_named_graph_2():
    ng1 = "urn:graph:gr1"
    ng2 = "urn:graph:gr2"
    m = Model()
    m.reads(gr1, format="ntriples", graph=ng1)
    m.reads(gr2, format="ntriples", graph=ng2)

    df = m.query("""
    SELECT * WHERE {
        GRAPH <urn:graph:gr1> {
            ?a ?b1 ?c1 .
        }
        GRAPH <urn:graph:gr2> {
            ?a ?b2 ?c2 .
        }
    }    
    """)
    assert df.height == 1
    assert df.get_column("c2")[0] == "B"

def test_detach_graph():
    ng1 = "urn:graph:gr1"
    ng2 = "urn:graph:gr2"
    m = Model()
    m.reads(gr1, format="ntriples", graph=ng1)
    m.reads(gr2, format="ntriples", graph=ng2)
    m2 = m.detach_graph(ng2, preserve_name=True)

    df = m.query("""
    SELECT * WHERE {
        GRAPH <urn:graph:gr2> {
            ?a ?b2 ?c2 .
        }
    }    
    """)
    assert df.height == 0

    df = m.query("""
    SELECT * WHERE {
        GRAPH <urn:graph:gr1> {
            ?a ?b1 ?c1 .
        }
    }    
    """)
    assert df.height == 1


    df2 = m2.query("""
    SELECT * WHERE {
        GRAPH <urn:graph:gr2> {
            ?a ?b2 ?c2 .
        }
    }    
    """)
    assert df2.height == 1
    assert df2.get_column("c2")[0] == "B"

    df2 = m2.query("""
    SELECT * WHERE {
        GRAPH <urn:graph:gr1> {
            ?a ?b1 ?c1 .
        }
    }    
    """)
    assert df2.height == 0

def test_detach_default_graph():
    ng1 = "urn:graph:gr1"
    m = Model()
    m.reads(gr1, format="ntriples", graph=ng1)
    m.reads(gr2, format="ntriples")
    m2 = m.detach_graph()

    df = m.query("""
    SELECT * WHERE {
        ?a ?b2 ?c2 .
    }    
    """)
    assert df.height == 0

    df = m.query("""
    SELECT * WHERE {
        GRAPH <urn:graph:gr1> {
            ?a ?b1 ?c1 .
        }
    }    
    """)
    assert df.height == 1


    df2 = m2.query("""
    SELECT * WHERE {
        ?a ?b2 ?c2 .
    }    
    """)
    assert df2.height == 1
    assert df2.get_column("c2")[0] == "B"

    df2 = m2.query("""
    SELECT * WHERE {
        GRAPH <urn:graph:gr1> {
            ?a ?b1 ?c1 .
        }
    }    
    """)
    assert df2.height == 0
    m.writes(format="ntriples")



import polars as pl
import pytest
import rdflib
from polars.testing import assert_frame_equal
import pathlib
from maplib import Model

pl.Config.set_fmt_str_lengths(300)

PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata" / "rdf_parser"


def test_issue_8():
    m = Model()
    m.read(TESTDATA_PATH / "date_panic.nt", format="ntriples")
    df = m.query("""SELECT ?c WHERE {?a ?b ?c}""", native_dataframe=True)
    expected = pl.from_repr(
        """
┌────────────┐
│ c          │
│ ---        │
│ date       │
╞════════════╡
│ 2035-01-23 │
└────────────┘
    """
    )
    assert_frame_equal(df, expected)



import polars as pl
import pytest
from polars.testing import assert_frame_equal

from maplib import BlankNode

from maplib import (
    Model,
    Prefix,
    Template,
    Argument,
    Parameter,
    Variable,
    RDFType,
    Triple,
    a,
)

pl.Config.set_fmt_str_lengths(200)


@pytest.fixture(scope="function")
def template() -> Template:
    pi = "https://github.com/DataTreehouse/maplib/pizza#"
    pi = Prefix(pi)

    p_var = Variable("p")
    c_var = Variable("c")
    ings_var = Variable("ings")

    template = Template(
        iri=pi.suf("PizzaTemplate"),
        parameters=[
            Parameter(variable=p_var, rdf_type=RDFType.IRI),
            Parameter(variable=c_var, rdf_type=RDFType.IRI),
            Parameter(variable=ings_var, rdf_type=RDFType.Nested(RDFType.IRI)),
        ],
        instances=[
            Triple(p_var, a, pi.suf("Pizza")),
            Triple(p_var, pi.suf("fromCountry"), c_var),
            Triple(p_var, pi.suf("hasBlank"), BlankNode("MyBlank")),
            Triple(
                p_var,
                pi.suf("hasIngredient"),
                Argument(term=ings_var, list_expand=True),
                list_expander="cross",
            ),
        ],
    )
    return template


@pytest.fixture(scope="function")
def pizzas_model(template: Template):

    pi = "https://github.com/DataTreehouse/maplib/pizza#"
    df = pl.DataFrame(
        {
            "p": [pi + "Hawaiian", pi + "Grandiosa"],
            "c": [pi + "CAN", pi + "NOR"],
            "ings": [[pi + "Pineapple", pi + "Ham"], [pi + "Pepper", pi + "Meat"]],
        }
    )
    # print(df)

    m = Model()
    m.map(template, df)
    hpizzas = """
    PREFIX pi:<https://github.com/DataTreehouse/maplib/pizza#>
    CONSTRUCT { ?p a pi:HeterodoxPizza } 
    WHERE {
        ?p a pi:Pizza .
        ?p pi:hasIngredient pi:Pineapple .
    }"""
    m.insert(hpizzas)
    return m


def test_simple_query_no_error(pizzas_model):
    res = pizzas_model.query(
        """
    PREFIX pi:<https://github.com/DataTreehouse/maplib/pizza#>

    SELECT ?p WHERE {
    ?p a pi:HeterodoxPizza
    }
    """
    )

    expected_df = pl.DataFrame(
        {"p": ["<https://github.com/DataTreehouse/maplib/pizza#Hawaiian>"]}
    )
    assert_frame_equal(res, expected_df)


def test_insert_new_thing(pizzas_model):
    hpizzas = """
    PREFIX pi:<https://github.com/DataTreehouse/maplib/pizza#>
    CONSTRUCT { ?p a pi:HeterodoxPizza2 } 
    WHERE {
        ?p a pi:Pizza .
        ?p pi:hasIngredient pi:Pineapple .
    }"""
    res1 = pizzas_model.insert(hpizzas)
    assert isinstance(res1, dict)
    assert len(res1) == 1
    assert res1["http://www.w3.org/1999/02/22-rdf-syntax-ns#type"].shape == (1, 2)
    res2 = pizzas_model.insert(hpizzas)
    assert len(res2) == 0


def test_insert_new_things(pizzas_model):
    hpizzas = """
    PREFIX pi:<https://github.com/DataTreehouse/maplib/pizza#>
    CONSTRUCT { 
        ?p a pi:HeterodoxPizza2 .
        ?p pi:abc pi:123 .
     } 
    WHERE {
        ?p a pi:Pizza .
    }"""
    res1 = pizzas_model.insert(hpizzas)
    assert isinstance(res1, dict)
    assert len(res1) == 2
    assert res1["http://www.w3.org/1999/02/22-rdf-syntax-ns#type"].shape == (2, 2)
    assert res1["https://github.com/DataTreehouse/maplib/pizza#abc"].shape == (2, 2)
    res2 = pizzas_model.insert(hpizzas)
    assert len(res2) == 0


def test_print_template(template: Template):
    s = str(template)
    assert (
        s
        == """<https://github.com/DataTreehouse/maplib/pizza#PizzaTemplate> [
     <http://ns.ottr.xyz/0.4/IRI> ?p, 
     <http://ns.ottr.xyz/0.4/IRI> ?c, 
     List<<http://ns.ottr.xyz/0.4/IRI>> ?ings ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(?p, <http://www.w3.org/1999/02/22-rdf-syntax-ns#type>, <https://github.com/DataTreehouse/maplib/pizza#Pizza>) ,
  <http://ns.ottr.xyz/0.4/Triple>(?p, <https://github.com/DataTreehouse/maplib/pizza#fromCountry>, ?c) ,
  <http://ns.ottr.xyz/0.4/Triple>(?p, <https://github.com/DataTreehouse/maplib/pizza#hasBlank>, _:MyBlank) ,
  cross | <http://ns.ottr.xyz/0.4/Triple>(?p, <https://github.com/DataTreehouse/maplib/pizza#hasIngredient>, ++ ?ings)
} . 
"""
    )



from typing import Dict, Optional
from .ordering import topological_sort
import polars as pl

from maplib import Model, Variable, RDFType, Parameter, Triple, IRI, Template


def generate_templates(m: Model, graph: Optional[str]) -> Dict[str, Template]:
    """Generate templates for instantiating the classes in an ontology

    :param m: The model where the ontology is stored. We mainly rely on rdfs:subClassOf, rdfs:range and rdfs:domain.
    :param graph: The named graph where the ontology is stored.

    :return A dictionary of templates for instantiating the classes in the ontology, where the keys are the class URIs.

    Usage example - note that it is important to add the templates to the Model you want to populate.
    >>> from maplib import Model, create_templates
    >>>
    >>> m_ont = Model()
    >>> m_ont.read("my_ontology.ttl")
    >>> templates = generate_templates(m_ont)
    >>> m = Model()
    >>> for t in templates.values():
    >>>     m.add_template(t)
    >>> m.map("https://example.net/MyClass", df)
    """

    properties = get_properties(m, graph=graph)
    properties_by_domain = {}
    properties_by_range = {}
    for r in properties.iter_rows(named=True):
        dom = r["domain"]
        if dom in properties_by_domain:
            properties_by_domain[dom].append(r)
        else:
            properties_by_domain[dom] = [r]

        ran = r["range"]
        if ran in properties_by_range:
            properties_by_range[ran].append(r)
        else:
            properties_by_range[ran] = [r]

    subclasses = get_subclasses(m, graph=graph)

    subclass_of = {}
    for r in (
        subclasses.group_by("child")
        .agg(pl.col("parent").alias("parents"))
        .iter_rows(named=True)
    ):
        subclass_of[r["child"]] = r["parents"]

    class_ordering = topological_sort(subclasses)

    templates_without_typing = generate_templates_without_typing(
        properties_by_domain, properties_by_range, class_ordering, subclass_of
    )
    templates_with_typing = generate_templates_with_typing(templates_without_typing)
    templates = {}
    for t, template in templates_without_typing.items():

        templates[t + "_notype"] = template
    for t, template in templates_with_typing.items():
        templates[t] = template

    return templates


def get_properties(m: Model, graph: Optional[str]) -> pl.DataFrame:
    properties = m.query(
        """
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
        PREFIX owl:  <http://www.w3.org/2002/07/owl#> 

        SELECT ?property ?property_type ?domain ?range WHERE { 
            ?property a ?property_type .
            ?property rdfs:domain ?domain .
            ?property rdfs:range ?range .
            FILTER(ISIRI(?domain) && ISIRI(?range))
        }
        """,
        native_dataframe=True,
        graph=graph,
    )
    pl.Config.set_fmt_str_lengths(100)
    properties_by_subclass_restriction = m.query(
        """
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
        PREFIX owl:  <http://www.w3.org/2002/07/owl#> 

        SELECT ?property ?property_type ?domain WHERE { 
            ?domain rdfs:subClassOf ?restr .
            ?restr a owl:Restriction .
            ?restr owl:onProperty ?property .
            #?property a ?property_type .
        }
        """,
        native_dataframe=True,
        graph=graph,
    )
    properties_by_subclass_restriction = (
        properties_by_subclass_restriction.with_columns(
            pl.lit("http://www.w3.org/2000/01/rdf-schema#ObjectType").alias(
                "property_type"
            ),
            pl.lit(None).cast(pl.String).alias("range"),
        )
    )
    if properties.height == 0:
        return properties_by_subclass_restriction
    elif properties_by_subclass_restriction.height == 0:
        return properties
    else:
        return properties.vstack(properties_by_subclass_restriction)

def get_subclasses(m: Model, graph: Optional[str]) -> pl.DataFrame:
    subclasses = m.query(
        """
        PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#> 
        SELECT ?child ?parent WHERE { 
            ?child rdfs:subClassOf ?parent .
            FILTER(ISIRI(?child) && ISIRI(?parent))
        }
        """,
        native_dataframe=True,
        graph=graph,
    )
    return subclasses


def uri_to_variable(uri: str) -> Variable:
    split = uri.split("/")
    name = split[-1].split("#")[-1]
    cleaned_name = name.replace("-", "_").replace(".", "_")
    return Variable(cleaned_name)


def generate_templates_without_typing(
    properties_by_domain,
    properties_by_range,
    class_ordering,
    subclass_of: Dict[str, str],
) -> Dict[str, Template]:
    templates = {}
    i = 0
    subj = get_subj_var()

    for c in class_ordering:
        i += 1

        parameters = []
        instances = []
        existing_preds = set()
        existing_varnames = set()

        # Check dupe!!
        subj_parameter = Parameter(
            variable=subj, optional=False, rdf_type=RDFType.IRI
        )
        parameters.append(subj_parameter)
        if c in properties_by_domain:
            for p in properties_by_domain[c]:
                if p["property"] in existing_preds:
                    # print("dupe: ", str(p))
                    continue
                existing_preds.add(p["property"])
                v = uri_to_variable(p["property"])
                existing_varnames.add(v.name)
                if p["property_type"] == "http://www.w3.org/2002/07/owl#ObjectProperty":
                    t = RDFType.IRI
                elif p["range"]:
                    t = RDFType.Literal(p["range"])
                else:
                    t = None
                param = Parameter(variable=v, optional=True, rdf_type=t)
                parameters.append(param)
                predicate = IRI(p["property"])
                triple = Triple(subj, predicate, v)
                instances.append(triple)
        if c in properties_by_range:
            for p in properties_by_range[c]:
                if p["property"] in existing_preds:
                    # print("dupe: ", str(p))
                    continue
                existing_preds.add(p["property"])
                v = uri_to_variable(p["property"])
                existing_varnames.add(v.name)
                t = RDFType.IRI
                param = Parameter(variable=v, optional=True, rdf_type=t)
                parameters.append(param)
                predicate = IRI(p["property"])
                triple = Triple(v, predicate, subj)
                instances.append(triple)
        c_tpl_iri = c + "_notype"
        if c in subclass_of:
            for sc in subclass_of[c]:
                if sc in templates:
                    sct = templates[sc]
                    variables = []
                    for p in sct.parameters:
                        if p.variable.name not in existing_varnames:
                            variables.append(p.variable)
                            if p.variable.name != "id":
                                parameters.append(p)
                        else:
                            variables.append(None)
                            # print(f"Duplicate variable: {str(p.variable.name)}")
                    instances.append(sct.instance(variables))

        tpl = Template(IRI(c_tpl_iri), parameters=parameters, instances=instances)
        templates[c] = tpl
    return templates


def generate_templates_with_typing(
    templates: Dict[str, Template],
) -> Dict[str, Template]:
    subj = get_subj_var()
    with_type_templates = {}
    for t, template in templates.items():
        instances = template.instances
        instances.append(
            Triple(subj, IRI("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"), IRI(t))
        )
        new_template = Template(
            iri=IRI(t), parameters=template.parameters, instances=instances
        )
        with_type_templates[t] = new_template
    return with_type_templates


def get_subj_var() -> Variable:
    return Variable("id")



from maplib import Model, IndexingOptions
import pytest
from polars.testing import assert_frame_equal
import polars as pl
from math import floor
import pathlib
import time

from pyparsing import stringStart

pl.Config.set_fmt_str_lengths(300)

PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata"


@pytest.fixture(scope="function", params=[True, False])
def windpower_model(request):
    print(request)
    instance_mapping = """
    @prefix tpl:<https://github.com/magbak/chrontext/templates#>.
    @prefix rds:<https://github.com/magbak/chrontext/rds_power#>.
    @prefix ct:<https://github.com/magbak/chrontext#>.
    @prefix rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    @prefix rdfs:<http://www.w3.org/2000/01/rdf-schema#>.
    
    tpl:Site [?SiteIRI, ?SiteName] :: {
        ottr:Triple(?SiteIRI, rdfs:label, ?SiteName),
        ottr:Triple(?SiteIRI, rdf:type, rds:Site)
        } .
    
    tpl:FunctionalAspect [?SourceIRI, ?TargetIRI, ?TargetAspectNodeIRI, ?Label] :: {
        ottr:Triple(?SourceIRI, rds:hasFunctionalAspect, ?TargetAspectNodeIRI),
        ottr:Triple(?TargetIRI, rds:hasFunctionalAspectNode, ?TargetAspectNodeIRI),
        ottr:Triple(?TargetAspectNodeIRI, rdfs:label, ?Label)
    } .
    
    tpl:RDSSystem [?SystemIRI, ottr:IRI ?RDSType, ?Label] :: {
        ottr:Triple(?SystemIRI, rdf:type, ?RDSType),
        ottr:Triple(?SystemIRI, rdfs:label, ?Label)
    } .
    
    tpl:StaticProperty [?ParentIRI, ottr:IRI ?ValueNodeIRI, ?Label, ?Value] :: {
        ottr:Triple(?ParentIRI, ct:hasStaticProperty, ?ValueNodeIRI),
        ottr:Triple(?ValueNodeIRI, rdfs:label, ?Label),
        ottr:Triple(?ValueNodeIRI, ct:hasStaticValue, ?Value)
    } .
    
    tpl:Timeseries [?ParentIRI, ottr:IRI ?TimeseriesNodeIRI, ?Label, ?ExternalId, ottr:IRI ?Datatype] :: {
        ottr:Triple(?ParentIRI, ct:hasTimeseries, ?TimeseriesNodeIRI),
        ottr:Triple(?TimeseriesNodeIRI, ct:hasExternalId, ?ExternalId),
        ottr:Triple(?TimeseriesNodeIRI, ct:hasDatatype, ?Datatype),
        ottr:Triple(?TimeseriesNodeIRI, rdfs:label, ?Label)
    } .
    """

    n = 40

    model = Model()
    model.add_template(instance_mapping)
    # Used as a prefix
    wpex = "https://github.com/magbak/chrontext/windpower_example#"
    rds = "https://github.com/magbak/chrontext/rds_power#"
    site_iris = [wpex + "Site" + str(i) for i in range(4)]
    sites = pl.DataFrame(
        {
            "SiteName": [
                "Wind Mountain",
                "Gale Valley",
                "Gusty Plains",
                "Breezy Field",
            ],
            "SiteIRI": site_iris,
        }
    )
    model.map("tpl:Site", sites)

    wind_turbine_iris = [wpex + "WindTurbine" + str(i) for i in range(1, n + 1)]
    wind_turbines = pl.DataFrame(
        {
            "RDSType": [rds + "A"] * n,
            "Label": ["Wind turbine " + str(i) for i in range(1, n + 1)],
            "SystemIRI": wind_turbine_iris,
        }
    )

    model.map("tpl:RDSSystem", wind_turbines)
    operating_external_ids = ["oper" + str(i) for i in range(1, n + 1)]
    operating = pl.DataFrame(
        {
            "Label": ["Operating"] * n,
            "ParentIRI": wind_turbine_iris,
            "ExternalId": operating_external_ids,
            "Datatype": ["http://www.w3.org/2001/XMLSchema#boolean"] * n,
        }
    )
    operating = operating.with_columns(
        pl.Series("TimeseriesNodeIRI", [wpex] * operating.height)
        + operating["ExternalId"]
    )
    model.map("tpl:Timeseries", operating)
    maximum_power_values = [
        [5_000_000, 10_000_000, 15_000_000][i % 3] for i in range(1, n + 1)
    ]
    maximum_power_node_iri = [
        wpex + "WindTurbineMaximumPower" + str(i) for i in range(1, n + 1)
    ]
    maximum_power = pl.DataFrame(
        {
            "Label": ["MaximumPower"] * n,
            "Value": maximum_power_values,
            "ParentIRI": wind_turbine_iris,
            "ValueNodeIRI": maximum_power_node_iri,
        }
    )
    model.map("tpl:StaticProperty", df=maximum_power)

    def add_aspect_labeling_by_source(df: pl.DataFrame, prefix: str) -> pl.DataFrame:
        label_df = df.group_by("SourceIRI", maintain_order=True).map_groups(
            lambda x: pl.DataFrame(
                {"Label": [prefix + str(i) for i in range(1, x.height + 1)]}
            )
        )
        df = df.with_columns(label_df["Label"])
        return df

    target_aspect_node_iri = [
        wpex + "WindTurbineFunctionalAspect" + str(i) for i in range(1, n + 1)
    ]

    site_has_wind_turbine = pl.DataFrame(
        {
            "SourceIRI": [
                site_iris[floor(i / (n / sites.height))] for i in range(0, n)
            ],
            "TargetIRI": wind_turbine_iris,
            "TargetAspectNodeIRI": target_aspect_node_iri,
        }
    )
    site_has_wind_turbine = add_aspect_labeling_by_source(site_has_wind_turbine, "A")
    model.map("tpl:FunctionalAspect", site_has_wind_turbine)

    generator_system_iris = [wpex + "GeneratorSystem" + str(i) for i in range(1, n + 1)]
    generator_systems = pl.DataFrame(
        {
            "SystemIRI": generator_system_iris,
            "Label": ["Generator system"] * n,
            "RDSType": [rds + "RA"] * n,
        }
    )
    model.map("tpl:RDSSystem", generator_systems)
    wind_turbine_has_generator_system = pl.DataFrame(
        {
            "SourceIRI": wind_turbine_iris,
            "TargetIRI": generator_system_iris,
            "TargetAspectNodeIRI": [
                wpex + "GeneratorSystemFunctionalAspect" + str(i)
                for i in range(1, n + 1)
            ],
        }
    )
    wind_turbine_has_generator_system = add_aspect_labeling_by_source(
        wind_turbine_has_generator_system, "RA"
    )
    model.map("tpl:FunctionalAspect", wind_turbine_has_generator_system)

    generator_iris = [wpex + "Generator" + str(i) for i in range(1, n + 1)]
    generators = pl.DataFrame(
        {
            "SystemIRI": generator_iris,
            "RDSType": [rds + "GAA"] * n,
            "Label": ["Generator"] * n,
        }
    )
    model.map("tpl:RDSSystem", generators)
    energy_production_external_ids = ["ep" + str(i) for i in range(1, n + 1)]
    energy_production = pl.DataFrame(
        {
            "Label": ["Production"] * n,
            "ParentIRI": generator_iris,
            "ExternalId": energy_production_external_ids,
            "Datatype": ["http://www.w3.org/2001/XMLSchema#double"] * n,
        }
    )
    energy_production = energy_production.with_columns(
        pl.Series("TimeseriesNodeIRI", [wpex] * energy_production.height)
        + energy_production["ExternalId"]
    )
    model.map("tpl:Timeseries", energy_production)
    generator_system_has_generator = pl.DataFrame(
        {
            "SourceIRI": generator_system_iris,
            "TargetIRI": generator_iris,
            "TargetAspectNodeIRI": [
                wpex + "GeneratorFunctionalAspect" + str(i) for i in range(1, n + 1)
            ],
        }
    )
    generator_system_has_generator = add_aspect_labeling_by_source(
        generator_system_has_generator, "GAA"
    )

    model.map("tpl:FunctionalAspect", generator_system_has_generator)

    wms_iris = [wpex + "WeatherMeasuringSystem" + str(i) for i in range(1, n + 1)]
    wms = pl.DataFrame(
        {
            "SystemIRI": wms_iris,
            "RDSType": [rds + "LE"] * n,
            "Label": ["Weather Measuring System"] * n,
        }
    )
    model.map("tpl:RDSSystem", wms)
    wind_speed_external_ids = ["wsp" + str(i) for i in range(1, n + 1)]
    wind_speed = pl.DataFrame(
        {
            "Label": ["Windspeed"] * n,
            "ParentIRI": wms_iris,
            "ExternalId": wind_speed_external_ids,
            "Datatype": ["http://www.w3.org/2001/XMLSchema#double"] * n,
        }
    )
    wind_speed = wind_speed.with_columns(
        pl.Series("TimeseriesNodeIRI", [wpex] * wind_speed.height)
        + wind_speed["ExternalId"]
    )

    model.map("tpl:Timeseries", wind_speed)
    wind_direction_external_ids = ["wdir" + str(i) for i in range(1, n + 1)]
    wind_direction = pl.DataFrame(
        {
            "Label": ["WindDirection"] * n,
            "ParentIRI": wms_iris,
            "ExternalId": wind_direction_external_ids,
            "Datatype": ["http://www.w3.org/2001/XMLSchema#double"] * n,
        }
    )
    wind_direction = wind_direction.with_columns(
        pl.Series("TimeseriesNodeIRI", [wpex] * wind_direction.height)
        + wind_direction["ExternalId"]
    )

    model.map("tpl:Timeseries", wind_direction)
    wind_turbine_has_wms = pl.DataFrame(
        {
            "SourceIRI": wind_turbine_iris,
            "TargetIRI": wms_iris,
            "TargetAspectNodeIRI": [
                wpex + "WMSFunctionalAspect" + str(i) for i in range(1, n + 1)
            ],
        }
    )
    wind_turbine_has_wms = add_aspect_labeling_by_source(wind_turbine_has_wms, "LE")

    model.map("tpl:FunctionalAspect", df=wind_turbine_has_wms)
    return model


@pytest.mark.parametrize("streaming", [True, False])
def test_simple_query(windpower_model, streaming):
    df = windpower_model.query(
        "SELECT ?a ?b WHERE {?a a ?b}", streaming=streaming
    ).sort(["a", "b"])
    filename = TESTDATA_PATH / "simple_query.csv"
    # df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(["a", "b"]).collect()
    pl.testing.assert_frame_equal(df, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_everything_from_subject_query(windpower_model, streaming):
    query = """
SELECT ?b ?c WHERE {
    <https://github.com/magbak/chrontext/windpower_example#Generator31> ?b ?c .
} ORDER BY ?b ?c"""
    windpower_model.query(query, streaming=streaming)

@pytest.mark.parametrize("streaming", [True, False])
def test_types(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?n ?t WHERE {
    ?n a ?t .
}"""
    df = windpower_model.query(query, streaming=streaming)
    print(df)
    assert df.height == 164



@pytest.mark.parametrize("streaming", [True, False])
def test_missing_predicate(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?n ?asp WHERE {
    ?n rds:hasFunctionalAspect ?asp .
}"""
    df = windpower_model.query(query, streaming=streaming)
    print(df)
    assert df.height == 160


@pytest.mark.parametrize("streaming", [True, False])
def test_missing_predicate2(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?wtur ?wtur_asp WHERE {
        ?wtur rds:hasFunctionalAspectNode ?wtur_asp .

}"""
    df = windpower_model.query(query, streaming=streaming)
    print(df)
    assert df.height == 160


@pytest.mark.parametrize("streaming", [True, False])
def test_property_path_bug(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?wtur ?wtur_asp WHERE {
        ?wtur rds:hasFunctionalAspectNode / rdfs:abc* ?wtur_asp .

}"""
    df = windpower_model.query(query, streaming=streaming)
    print(df)
    assert df.height == 160




@pytest.mark.parametrize("streaming", [True, False])
def test_larger_query_simplified(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label ?wtur_label WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site rds:hasFunctionalAspect ?wtur_asp .
    ?wtur_asp rdfs:label ?wtur_label .
    ?wtur rds:hasFunctionalAspectNode ?wtur_asp .
    FILTER(?wtur_label = "A1" && ?site_label = "Wind Mountain") .
}"""
    df = windpower_model.query(query, streaming=streaming)
    print(df)


@pytest.mark.parametrize("streaming", [True, False])
def test_larger_query(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label ?wtur_label ?ts ?ts_label WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site rds:hasFunctionalAspect ?wtur_asp .
    ?wtur_asp rdfs:label ?wtur_label .
    ?wtur rds:hasFunctionalAspectNode ?wtur_asp .
    ?wtur rds:hasFunctionalAspect ?gensys_asp .
    ?wtur a rds:A .
    ?gensys rds:hasFunctionalAspectNode ?gensys_asp .
    ?gensys a rds:RA .
    ?gensys rds:hasFunctionalAspect ?generator_asp .
    ?generator rds:hasFunctionalAspectNode ?generator_asp .
    ?generator a rds:GAA .
    ?generator ct:hasTimeseries ?ts .
    ?ts rdfs:label ?ts_label .
    FILTER(?wtur_label = "A1" && ?site_label = "Wind Mountain") .
}"""
    by = ["site_label", "wtur_label", "ts", "ts_label"]
    df = windpower_model.query(query, streaming=streaming).sort(by)
    filename = TESTDATA_PATH / "larger_query.csv"
    # df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    # Important assert for checking indexing by object string
    assert_frame_equal(df, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_bad_pushdown(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label ?ts ?ts_label WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site rds:hasFunctionalAspect ?wtur_asp .
    ?wtur_asp rdfs:label 1 .
    ?wtur rds:hasFunctionalAspectNode ?wtur_asp .
    ?wtur rds:hasFunctionalAspect ?gensys_asp .
    ?wtur a rds:A .
    ?gensys rds:hasFunctionalAspectNode ?gensys_asp .
    ?gensys a rds:RA .
    ?gensys rds:hasFunctionalAspect ?generator_asp .
    ?generator rds:hasFunctionalAspectNode ?generator_asp .
    ?generator a rds:GAA .
    ?generator ct:hasTimeseries ?ts .
    ?ts rdfs:label ?ts_label .
    FILTER(?site_label = "Wind Mountain") .
}"""
    df = windpower_model.query(query, streaming=streaming)
    assert df.shape == (0, 3)


@pytest.mark.parametrize("streaming", [True, False])
def test_larger_ordered_query(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label ?wtur_label ?ts ?ts_label WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site rds:hasFunctionalAspect ?wtur_asp .
    ?wtur_asp rdfs:label ?wtur_label .
    ?wtur rds:hasFunctionalAspectNode ?wtur_asp .
    ?wtur rds:hasFunctionalAspect ?gensys_asp .
    ?wtur a rds:A .
    ?gensys rds:hasFunctionalAspectNode ?gensys_asp .
    ?gensys a rds:RA .
    ?gensys rds:hasFunctionalAspect ?generator_asp .
    ?generator rds:hasFunctionalAspectNode ?generator_asp .
    ?generator a rds:GAA .
    ?generator ct:hasTimeseries ?ts .
    ?ts rdfs:label ?ts_label .
} ORDER BY ?site_label ?wtur_label ?ts ?ts_label"""
    df = windpower_model.query(query, streaming=streaming)
    filename = TESTDATA_PATH / "larger_ordered_query.csv"
    # df.write_csv(filename)
    expected_df = pl.scan_csv(filename).collect()
    assert_frame_equal(df, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_larger_ordered_query_debug(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label ?wtur_label ?ts ?ts_label WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site rds:hasFunctionalAspect ?wtur_asp .
    ?wtur_asp rdfs:label ?wtur_label .
    ?wtur rds:hasFunctionalAspectNode ?wtur_asp .
    ?wtur rds:hasFunctionalAspect ?gensys_asp .
    ?wtur a rds:A .
    ?gensys rds:hasFunctionalAspectNode ?gensys_asp .
    ?gensys a rds:RA .
    #This one bad:
    ?gensys rds:hasFeunctionalAspect ?generator_asp .
    ?generator rds:hasFunctionalAspectNode ?generator_asp .
    ?generator a rds:GAA .
    ?generator ct:hasTimeseries ?ts .
    ?ts rdfs:label ?ts_label .
} ORDER BY ?site_label ?wtur_label ?ts ?ts_label"""
    sm = windpower_model.query(query, streaming=streaming, debug=True)
    print(sm.debug)
    assert "hasFeunc" in sm.debug


@pytest.mark.parametrize("streaming", [True, False])
def test_larger_ordered_query_debug_badfilter(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label ?wtur_label ?ts ?ts_label WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site rds:hasFunctionalAspect ?wtur_asp .
    ?wtur_asp rdfs:label ?wtur_label .
    ?wtur rds:hasFunctionalAspectNode ?wtur_asp .
    ?wtur rds:hasFunctionalAspect ?gensys_asp .
    ?wtur a rds:A .
    ?gensys rds:hasFunctionalAspectNode ?gensys_asp .
    ?gensys a rds:RA .
    ?gensys rds:hasFunctionalAspect ?generator_asp .
    ?generator rds:hasFunctionalAspectNode ?generator_asp .
    ?generator a rds:GAA .
    ?generator ct:hasTimeseries ?ts .
    ?ts rdfs:label ?ts_label .
    FILTER(?ts_label = "HALLO")
} ORDER BY ?site_label ?wtur_label ?ts ?ts_label"""
    sm = windpower_model.query(query, streaming=streaming, debug=True)
    print(sm.debug)
    assert "HALLO" in sm.debug

@pytest.mark.parametrize("streaming", [True, False])
def test_larger_ordered_query_debug_but_no_error(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label ?wtur_label ?ts ?ts_label WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site rds:hasFunctionalAspect ?wtur_asp .
    ?wtur_asp rdfs:label ?wtur_label .
    ?wtur rds:hasFunctionalAspectNode ?wtur_asp .
    ?wtur rds:hasFunctionalAspect ?gensys_asp .
    ?wtur a rds:A .
    ?gensys rds:hasFunctionalAspectNode ?gensys_asp .
    ?gensys a rds:RA .
    ?gensys rds:hasFunctionalAspect ?generator_asp .
    ?generator rds:hasFunctionalAspectNode ?generator_asp .
    ?generator a rds:GAA .
    ?generator ct:hasTimeseries ?ts .
    ?ts rdfs:label ?ts_label .
} ORDER BY ?site_label ?wtur_label ?ts ?ts_label"""
    sm = windpower_model.query(query, streaming=streaming, debug=True)
    assert "has at least one" in sm.debug


@pytest.mark.parametrize("streaming", [True, False])
def test_simple_property_path_query(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label ?node WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site rds:hasFunctionalAspect / ^rds:hasFunctionalAspectNode ?node .
}"""
    by = ["site_label", "node"]
    df = windpower_model.query(query, streaming=streaming).sort(by)
    filename = TESTDATA_PATH / "simple_property_path_query.csv"
    # df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    pl.testing.assert_frame_equal(df, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_iterated_property_path_query(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label ?node WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site (rds:hasFunctionalAspect / ^rds:hasFunctionalAspectNode)+ ?node .
}"""
    by = ["site_label", "node"]
    df = windpower_model.query(query, streaming=streaming).sort(by)
    filename = TESTDATA_PATH / "iterated_property_path_query.csv"
    # df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    pl.testing.assert_frame_equal(df, expected_df)

@pytest.mark.parametrize("streaming", [True, False])
def test_iterated_property_path_constant_object_query(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site (rds:hasFunctionalAspect / ^rds:hasFunctionalAspectNode)+ <https://github.com/magbak/chrontext/windpower_example#Generator40> .
}"""
    by = ["site_label"]
    df = windpower_model.query(query, streaming=streaming).sort(by)
    filename = TESTDATA_PATH / "iterated_property_path_constant_object_query.csv"
    # df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    pl.testing.assert_frame_equal(df, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_iterated_property_path_constant_subject_query(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?node WHERE {
    <https://github.com/magbak/chrontext/windpower_example#Site3> (rds:hasFunctionalAspect / ^rds:hasFunctionalAspectNode)+ ?node .
}"""
    by = ["node"]
    start = time.time()
    df = windpower_model.query(query, streaming=streaming).sort(by)
    end = time.time()
    print(f"Took {round(end-start, 3)}")
    filename = TESTDATA_PATH / "iterated_property_path_constant_subject_query.csv"
    # df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    pl.testing.assert_frame_equal(df, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_iterated_property_path_query_with_bug(windpower_model, streaming):
    query = """PREFIX xsd:<http://www.w3.org/2001/XMLSchema#>
PREFIX ct:<https://github.com/magbak/chrontext#>
PREFIX wp:<https://github.com/magbak/chrontext/windpower_example#>
PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf:<http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rds:<https://github.com/magbak/chrontext/rds_power#>
SELECT ?site_label ?node_label WHERE {
    ?site a rds:Site .
    ?site rdfs:label ?site_label .
    ?site (rds:hasFunctionalAspect / ^rds:hasFunctionalAspectNode)+ ?node .
    ?node rdfs:label ?node_label .
}"""
    by = ["site_label", "node_label"]
    start = time.time()
    df = windpower_model.query(query, streaming=streaming).sort(by)
    end = time.time()
    print(f"Took {round(end-start, 3)}")
    filename = TESTDATA_PATH / "iterated_property_path_query_with_bug.csv"
    # df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(by).collect()
    assert_frame_equal(df, expected_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_simple_construct_query(windpower_model, streaming):
    dfs = windpower_model.query(
        """
    PREFIX ct:<https://github.com/magbak/chrontext#>
    CONSTRUCT {
    ?a a ct:something.
    ?b a ct:nothing. 
    } WHERE {?a a ?b}""",
        streaming=streaming,
    )
    something = dfs[0].sort(["subject", "object"])
    nothing = dfs[1].sort(["subject", "object"])
    filename_something = TESTDATA_PATH / "simple_construct_query_something.csv"
    # something.write_csv(filename_something)
    filename_nothing = TESTDATA_PATH / "simple_construct_query_nothing.csv"
    # nothing.write_csv(filename_nothing)
    expected_something_df = (
        pl.scan_csv(filename_something).sort(["subject", "object"]).collect()
    )
    assert_frame_equal(something, expected_something_df)
    expected_nothing_df = (
        pl.scan_csv(filename_nothing).sort(["subject", "object"]).collect()
    )
    assert_frame_equal(nothing, expected_nothing_df)


# No idempotency means no streaming parameter.
def test_simple_insert_construct_query(windpower_model):
    windpower_model.insert(
        """
    PREFIX ct:<https://github.com/magbak/chrontext#>
    CONSTRUCT {
    ?a a ct:somethingTestit.
    ?b a ct:nothingTestit. 
    } WHERE {?a a ?b}""",
    )

    something = windpower_model.query(
        """
        PREFIX ct:<https://github.com/magbak/chrontext#>
        SELECT ?a
        WHERE {
        ?a a ct:somethingTestit .
        }
    """,
    ).sort(["a"])
    nothing = windpower_model.query(
        """
            PREFIX ct:<https://github.com/magbak/chrontext#>
            SELECT ?a
            WHERE {
            ?a a ct:nothingTestit .
            }
        """,
    ).sort(["a"])
    filename_something = TESTDATA_PATH / "simple_insert_query_something.csv"
    # something.write_csv(filename_something)
    filename_nothing = TESTDATA_PATH / "simple_insert_query_nothing.csv"
    # nothing.write_csv(filename_nothing)
    expected_something_df = pl.scan_csv(filename_something).sort(["a"]).collect()
    assert_frame_equal(something, expected_something_df)
    expected_nothing_df = pl.scan_csv(filename_nothing).sort(["a"]).collect()
    assert_frame_equal(nothing, expected_nothing_df)


@pytest.mark.parametrize("streaming", [True, False])
def test_str_functions(windpower_model, streaming):
    pl.Config.set_tbl_rows(20)
    df = windpower_model.query(
        """
        PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
        SELECT 
            ?label
            (STRSTARTS(?label, "Wind turbine 3") AS ?strStarts)
            (STRENDS(?label, "11") AS ?strEnds)
            (CONTAINS(?label, "ind turbine 24") AS ?contains)
            (STRBEFORE(?label, "urbine 24") AS ?strBefore)
            (STRAFTER(?label, "Wind t") AS ?strAfter)
        WHERE {
        ?a rdfs:label ?label .
        }
        """,
        streaming=streaming,
    )
    df = df.filter(pl.col("strStarts") | pl.col("strEnds") | pl.col("contains")).sort(
        ["label"]
    )
    filename = TESTDATA_PATH / "stringfuncs.csv"
    # df.write_csv(filename)
    expected_df = pl.scan_csv(filename).sort(["label"]).collect()
    assert_frame_equal(df, expected_df)



import polars as pl
import pytest

from maplib import Model, RDFType, xsd
from polars.testing import assert_frame_equal
import polars as pl
import pathlib

pl.Config.set_fmt_str_lengths(200)
PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata"


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_equals(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    FILTER(?a = :hello2)
    }
    """,
        streaming=streaming,
    )
    assert_frame_equal(df, pl.DataFrame({"a": ["<http://example.net/hello2>"]}))

@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_numerical_equals(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b WHERE {
    {VALUES (?a) { ("1"^^xsd:int) }}
    {VALUES (?b) { ("1"^^xsd:integer) }}
    FILTER(?a = ?b)
    }
    """,
        streaming=streaming,
    )
    assert df.shape == (0, 2)


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_numerical_cast_real_equals(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b WHERE {
    {VALUES (?a) { ("1"^^xsd:int) }}
    {VALUES (?b) { ("1"^^xsd:integer) }}
    FILTER(xsd:double(?a) = xsd:double(?b))
    }
    """,
        streaming=streaming,
    )
    assert df.shape == (1, 2)


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_numerical_cast_integer_equals(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b WHERE {
        {VALUES (?a) { ("1"^^xsd:int) }}
        {VALUES (?b) { ("1"^^xsd:integer) }}
    FILTER(xsd:integer(?a) = xsd:integer(?b))
    }
    """,
        streaming=streaming,
    )
    assert df.shape == (1, 2)


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_numerical_cast_string_equals(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b WHERE {
    VALUES (?a) { ("1"^^xsd:int)}
    VALUES (?b) { ("1"^^xsd:integer)} 
    FILTER(xsd:string(?a) = xsd:string(?b))
    }
    """,
        streaming=streaming,
    )
    assert df.shape == (1, 2)


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_strlen(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a (STRLEN(?a) AS ?len) WHERE {
        VALUES (?a) { ("123"^^xsd:int) ("Hello") ("") ("ABC"^^xsd:string)}
    } 
    """,
        streaming=streaming,
    )
    assert df.get_column("len").sum() == 11


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_uri_cast_string_equals(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a  WHERE {
        { VALUES (?a) { ("1"^^xsd:int) }} UNION { VALUES (?a) { (:something) }}
        FILTER(xsd:string(?a) = "http://example.net/something")
    }
    """,
        streaming=streaming,
    )
    assert df.shape == (1, 1)


@pytest.mark.parametrize("streaming", [True, False])
def test_filter_uri_cast_string_equals(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a  WHERE {
        VALUES (?a) { (:something) }
        FILTER(xsd:string(?a) = "http://example.net/something")
    }
    """,
        streaming=streaming,
    )
    assert df.shape == (1, 1)


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_equals_with_datatypes(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    FILTER(?a = :hello2)
    }
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {"a": RDFType.IRI}
    assert_frame_equal(
        sm.mappings, pl.DataFrame({"a": ["<http://example.net/hello2>"]})
    )


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_value_different_types_filter_equals_with_datatypes(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) ("string") (1) (true) }}
    FILTER(?a = true)
    }
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {"a": RDFType.Literal(xsd.boolean)}
    assert_frame_equal(sm.mappings, pl.DataFrame({"a": [True]}))


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_value_different_types_filter_isiri_or_equals_with_datatypes(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) ("string") (1) (true) }}
    FILTER(isIri(?a) || ?a = 1)
    }
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi([RDFType.IRI, RDFType.Literal(xsd.integer)])
    }
    assert_frame_equal(
        sm.mappings,
        pl.DataFrame(
            {
                "a": [
                    "<http://example.net/hello2>",
                    '"1"^^<http://www.w3.org/2001/XMLSchema#integer>',
                ]
            }
        ),
    )


@pytest.mark.parametrize("streaming", [True, False])
def test_coalesce_both_multi_same_types(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    {VALUES (?a) { (:hello1) ("string1") (1) (true) }}
    OPTIONAL {
        VALUES (?b) { (:hello2) ("string2") (1) (false) }
        FILTER(isIri(?a))
    }
    }
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    print(sm.mappings)
    print(sm.rdf_types)
    assert sm.mappings.get_column("c").is_null().sum() == 0


@pytest.mark.parametrize("streaming", [True, False])
def test_coalesce_both_multi_different_types(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    {VALUES (?a) { (:hello1) ("string1") }}
    OPTIONAL {
        VALUES (?b) { (:hello2) ("string2"@en) (1) (false) }
        FILTER(isIri(?a))
    }
    }
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.mappings.get_column("c").is_null().sum() == 0


@pytest.mark.parametrize("streaming", [True, False])
def test_coalesce_both_multi_different_types_rhs_less(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    VALUES (?a) { (:hello1) ("string1") (1)}
    OPTIONAL {
        VALUES (?b) { (:hello2) ("string2") }
        FILTER(isIri(?a))
    }
    }
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.mappings.get_column("c").is_null().sum() == 0


@pytest.mark.parametrize("streaming", [True, False])
def test_coalesce_only_lhs_multi(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    VALUES (?a) { (:hello1) ("string1") (1)}
    OPTIONAL {
        VALUES (?b) { (:hello2) (:hello3) }
        FILTER(isIri(?a))
    }
    }
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.mappings.get_column("c").is_null().sum() == 0


@pytest.mark.parametrize("streaming", [True, False])
def test_coalesce_only_rhs_multi(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    VALUES (?a) { (:hello1) (:hello3) (:hello4)}
    OPTIONAL {
        VALUES (?b) { (:hello2) ("string2"@no) ("string3") }
        FILTER(?a = :hello1)
    }
    }
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.mappings.get_column("c").is_null().sum() == 0


@pytest.mark.parametrize("streaming", [True, False])
def test_div_by_unbound_is_unbound(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (?a / ?b AS ?c) WHERE {
    VALUES (?a) { (1) (2) }
    OPTIONAL {
        VALUES (?b) { (1.5) (4.5) }
        FILTER(?a = 1)
    }
    }
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.mappings.height == 3
    assert sm.mappings.get_column("c").is_null().sum() == 1


@pytest.mark.parametrize("streaming", [True, False])
def test_coalesce_no_multi(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    VALUES (?a) { (:hello1) (:hello2) (:hello3)}
    OPTIONAL {
        VALUES (?b) { (:hello5) (:hello6) }
        FILTER(?a = :hello1)
    }
    }
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {"c": RDFType.IRI, "b": RDFType.IRI, "a": RDFType.IRI}
    assert sm.mappings.get_column("c").is_null().sum() == 0


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_equals_mirrored(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    FILTER(:hello2 = ?a)
    }
    """,
        streaming=streaming,
    )
    assert_frame_equal(df, pl.DataFrame({"a": ["<http://example.net/hello2>"]}))


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_equals_two_multi(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    {VALUES (?b) { (:hello2) }} UNION {VALUES (?b) { (1) }}
    FILTER(?b = ?a)
    }
    """,
        streaming=streaming,
    )
    assert_frame_equal(df, pl.DataFrame({"a": ["<http://example.net/hello2>"]}))


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_is_in(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
        FILTER(?a in (:hello2, "abc"))
    }
    """,
        streaming=streaming,
    )
    print(df)
    assert_frame_equal(df, pl.DataFrame({"a": ["<http://example.net/hello2>"]}))


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_comparison(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    VALUES (?a) { (1) (3.0) (89) }
    FILTER(?a > 1.4)
    } ORDER BY ?a
    """,
        streaming=streaming,
    )
    expected = pl.from_repr(
        """
┌───────────────────────────────────────────────────┐
│ a                                                 │
│ ---                                               │
│ str                                               │
╞═══════════════════════════════════════════════════╡
│ "3.0"^^<http://www.w3.org/2001/XMLSchema#decimal> │
│ "89"^^<http://www.w3.org/2001/XMLSchema#integer>  │
└───────────────────────────────────────────────────┘
    """
    )
    assert_frame_equal(df, expected)


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_incompatible_comparison(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    VALUES (?a) { (1) (3.0) (89) ("String") ("2021-01-01T08:00:00"^^xsd:dateTime) }
    FILTER(?a > 1.4)
    } ORDER BY ?a
    """,
        streaming=streaming,
    )
    expected = pl.from_repr(
        """
┌───────────────────────────────────────────────────┐
│ a                                                 │
│ ---                                               │
│ str                                               │
╞═══════════════════════════════════════════════════╡
│ "3.0"^^<http://www.w3.org/2001/XMLSchema#decimal> │
│ "89"^^<http://www.w3.org/2001/XMLSchema#integer>  │
└───────────────────────────────────────────────────┘
    """
    )
    assert_frame_equal(df, expected)


@pytest.mark.parametrize("streaming", [True, False])
def test_parse_strange_year(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
        VALUES (?a) { ("9999-01-01T08:00:00"^^xsd:dateTime) }
    } 
    """,
        streaming=streaming,
    )
    expected = pl.from_repr(
        """
┌─────────────────────┐
│ a                   │
│ ---                 │
│ datetime[μs, UTC]   │
╞═════════════════════╡
│ 9999-01-01 08:00:00 │
└─────────────────────┘
    """
    )
    assert_frame_equal(df, expected)


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_incompatible_datetime_comparison(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    VALUES (?a) { (1) (3.0) (89) ("String") ("2021-01-01T08:00:00"^^xsd:dateTime) }
    FILTER(?a > "2021-01-01T07:59:59"^^xsd:dateTime)
    } ORDER BY ?a
    """,
        streaming=streaming,
    )
    expected = pl.from_repr(
        """
┌─────────────────────┐
│ a                   │
│ ---                 │
│ datetime[μs, UTC]   │
╞═════════════════════╡
│ 2021-01-01 08:00:00 │
└─────────────────────┘
    """
    )
    assert_frame_equal(df, expected)


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_incompatible_many_comparison(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b WHERE {
    VALUES (?a) { (1) (3.0) (89) ("String") ("2021-01-01T08:00:00"^^xsd:dateTime) }
    VALUES (?b) { (2) (2.0) (90) ("AString") ("2021-01-01T08:00:01"^^xsd:dateTime) }
    FILTER(?a < ?b)
    } ORDER BY ?a ?b
    """,
        streaming=streaming,
    )
    f = TESTDATA_PATH / "multi_many_comp.csv"
    #df.write_csv(f)
    expected = pl.read_csv(f)
    assert_frame_equal(df, expected)


@pytest.mark.parametrize("streaming", [True, False])
def test_multi_concat(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b ?c WHERE {
    VALUES (?a) { (1) (3.0) (89) ("String") ("2021-01-01T08:00:00"^^xsd:dateTime) }
    VALUES (?b) { (2) (2.0) (90) ("AString") ("2021-01-01T08:00:01"^^xsd:dateTime) }
    BIND(CONCAT(STR(?a),"_", STR(?b)) AS ?c)
    } ORDER BY ?a ?b
    """,
        streaming=streaming,
    )
    f = TESTDATA_PATH / "multi_concat.csv"
    #df.write_csv(f)
    expected = pl.read_csv(f)
    assert_frame_equal(df, expected)

@pytest.mark.parametrize("streaming", [True, False])
def test_multi_concat_count(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT (COUNT(*) AS ?c) WHERE {
    VALUES (?a) { (1) (3.0) (89) ("String") ("2021-01-01T08:00:00"^^xsd:dateTime) }
    VALUES (?b) { (2) (2.0) (90) ("AString") ("2021-01-01T08:00:01"^^xsd:dateTime) }
    } 
    """,
        streaming=streaming,
    )
    assert df.shape == (1,1)
    assert df.get_column("c")[0] == 25


@pytest.mark.parametrize("streaming", [True, False])
def test_strlang(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT * WHERE {
    BIND(STRLANG("a", "en") AS ?a)
    } 
    """,
        streaming=streaming,
    )
    assert df.height == 1
    assert df.get_column("a")[0] == "\"a\"@en"

@pytest.mark.parametrize("streaming", [True, False])
def test_multi_filter_incompatible_datetimestamp_comparison(streaming):
    m = Model()
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    VALUES (?a) { (1) (3.0) (89) ("String") ("2021-01-01T08:00:00+00:00"^^xsd:dateTime) }
    FILTER(?a > "2021-01-01T07:59:59+00:00"^^xsd:dateTimeStamp)
    } ORDER BY ?a
    """,
        streaming=streaming,
    )
    expected = pl.from_repr(
        """
┌─────────────────────┐
│ a                   │
│ ---                 │
│ datetime[μs, UTC]   │
╞═════════════════════╡
│ 2021-01-01 08:00:00 │
└─────────────────────┘
    """
    )
    assert_frame_equal(df, expected)


@pytest.mark.parametrize("streaming", [True, False])
def test_generate_uuids(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?uuid WHERE {
    VALUES (?a) { (1) (2) (3) }
    BIND(uuid() as ?uuid)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
        "uuid": RDFType.IRI,
    }
    assert sm.mappings.height == 3
    df = sm.mappings.with_columns(
        pl.col("uuid").str.starts_with("<urn:uuid:").alias("startswithcorrect")
    )
    assert df.get_column("startswithcorrect").sum() == 3
    assert df.get_column("uuid").unique().len() == 3


@pytest.mark.parametrize("streaming", [True, False])
def test_generate_iri_all_strings(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?iri WHERE {
    VALUES (?a) { ("urn:abc:123") ("http://www.w3.org/2001/XMLSchema#integer") }
    BIND(IRI(?a) as ?iri)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
        "iri": RDFType.IRI,
    }
    expected = pl.from_repr(
        """
┌─────────────────────────────────────────────────┬────────────────────────────────────────────┐
│ a                                               ┆ iri                                        │
│ ---                                             ┆ ---                                        │
│ str                                             ┆ str                                        │
╞═════════════════════════════════════════════════╪════════════════════════════════════════════╡
│ http://www.w3.org/2001/XMLSchema#integer        ┆ <http://www.w3.org/2001/XMLSchema#integer> │
│ urn:abc:123                                     ┆ <urn:abc:123>                              │
└─────────────────────────────────────────────────┴────────────────────────────────────────────┘
    """
    )
    assert_frame_equal(expected, sm.mappings)


@pytest.mark.parametrize("streaming", [True, False])
def test_generate_iri(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?iri WHERE {
    VALUES (?a) { ("urn:abc:123") ("http://www.w3.org/2001/XMLSchema#integer") (3) (<http://www.w3.org/2001/XMLSchema#abc>) }
    BIND(IRI(?a) as ?iri)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi(
            [
                RDFType.IRI,
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
        "iri": RDFType.IRI,
    }
    expected = pl.from_repr(
        """
┌─────────────────────────────────────────────────┬────────────────────────────────────────────┐
│ a                                               ┆ iri                                        │
│ ---                                             ┆ ---                                        │
│ str                                             ┆ str                                        │
╞═════════════════════════════════════════════════╪════════════════════════════════════════════╡
│ <http://www.w3.org/2001/XMLSchema#abc>          ┆ <http://www.w3.org/2001/XMLSchema#abc>     │
│ "3"^^<http://www.w3.org/2001/XMLSchema#integer> ┆ null                                       │
│ "http://www.w3.org/2001/XMLSchema#integer"      ┆ <http://www.w3.org/2001/XMLSchema#integer> │
│ "urn:abc:123"                                   ┆ <urn:abc:123>                              │
└─────────────────────────────────────────────────┴────────────────────────────────────────────┘
    """
    )
    assert_frame_equal(expected, sm.mappings)


@pytest.mark.parametrize("streaming", [True, False])
def test_generate_str_uuids(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?struuid WHERE {
    VALUES (?a) { (1) (2) (3) }
    BIND(struuid() as ?struuid)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
        "struuid": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
    }
    assert sm.mappings.height == 3
    assert sm.mappings.get_column("struuid").unique().len() == 3


@pytest.mark.parametrize("streaming", [True, False])
def test_replace_single(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?replace WHERE {
    VALUES (?a) { ("abcabc") ("ab") ("bb") }
    BIND(REPLACE(?a, "ab", "ba") as ?replace)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
        "replace": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
    }
    assert sm.mappings.height == 3
    assert sm.mappings.get_column("replace").to_list() == ["ba", "bacbac", "bb"]


@pytest.mark.parametrize("streaming", [True, False])
def test_case_single(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?ucase ?lcase WHERE {
    VALUES (?a) { ("abcaAc") ("ab") ("bb") }
    BIND(UCASE(?a) as ?ucase)
    BIND(LCASE(?a) as ?lcase)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
        "ucase": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
        "lcase": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
    }
    assert sm.mappings.height == 3
    assert sm.mappings.get_column("ucase").to_list() == ["AB", "ABCAAC", "BB"]
    assert sm.mappings.get_column("lcase").to_list() == ["ab", "abcaac", "bb"]


@pytest.mark.parametrize("streaming", [True, False])
def test_mul_types_single(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT * WHERE {
    VALUES (?a) { ("dummy") }
    BIND("0.1"^^xsd:float * "0.1"^^xsd:decimal AS ?n1)
    BIND("0.1"^^xsd:double * "0.1"^^xsd:decimal AS ?n2)
    BIND("0.1"^^xsd:decimal * "0.1"^^xsd:decimal AS ?n3)
    BIND("0.1"^^xsd:float * "0.1"^^xsd:float AS ?n4)
    BIND("0.1"^^xsd:double * "0.1"^^xsd:double AS ?n5)
    BIND("21"^^xsd:integer * "0.1"^^xsd:decimal AS ?n6)
    #BIND("true"^^xsd:boolean * "0.1"^^xsd:decimal AS ?n7)
    BIND("1"^^xsd:integer * "2"^^xsd:long AS ?n8)
    BIND("2"^^xsd:long * "2"^^xsd:long AS ?n9)
    BIND("2"^^xsd:int * "2"^^xsd:int AS ?n10)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
        "n1": RDFType.Literal("http://www.w3.org/2001/XMLSchema#decimal"),
        "n2": RDFType.Literal("http://www.w3.org/2001/XMLSchema#decimal"),
        "n3": RDFType.Literal("http://www.w3.org/2001/XMLSchema#decimal"),
        "n4": RDFType.Literal("http://www.w3.org/2001/XMLSchema#float"),
        "n5": RDFType.Literal("http://www.w3.org/2001/XMLSchema#double"),
        "n6": RDFType.Literal("http://www.w3.org/2001/XMLSchema#decimal"),
        "n8": RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
        "n9": RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
        "n10": RDFType.Literal("http://www.w3.org/2001/XMLSchema#int"),
    }


@pytest.mark.parametrize("streaming", [True, False])
def test_substr_single(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?sub1 ?sub2 WHERE {
    VALUES (?a) { ("abcaAc") ("ab") ("bb") }
    BIND(SUBSTR(?a,3) as ?sub1)
    BIND(SUBSTR(?a,0,3) as ?sub2)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
        "sub1": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
        "sub2": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
    }
    assert sm.mappings.height == 3
    assert sm.mappings.get_column("sub1").to_list() == ["", "aAc", ""]
    assert sm.mappings.get_column("sub2").to_list() == ["ab", "abc", "bb"]


@pytest.mark.parametrize("streaming", [True, False])
def test_case_single_lang(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?ucase ?lcase WHERE {
    VALUES (?a) { ("abcaAc"@en) ("ab"@no) ("bb"@se) }
    BIND(UCASE(?a) as ?ucase)
    BIND(LCASE(?a) as ?lcase)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"),
        "ucase": RDFType.Literal(
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
        ),
        "lcase": RDFType.Literal(
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
        ),
    }
    assert sm.mappings.height == 3
    assert sm.mappings.get_column("ucase").to_list() == [
        '"AB"@no',
        '"ABCAAC"@en',
        '"BB"@se',
    ]
    assert sm.mappings.get_column("lcase").to_list() == [
        '"ab"@no',
        '"abcaac"@en',
        '"bb"@se',
    ]


@pytest.mark.parametrize("streaming", [True, False])
def test_substr_multi_type_with_only_lang(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?sub1 ?sub2 WHERE {
    VALUES (?a) { ("abcaAc"@en) ("ab"@no) ("bb"@se) (1) }
    BIND(SUBSTR(?a, 2) as ?sub1)
    BIND(SUBSTR(?a, 2, 1) as ?sub2)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi(
            [
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
            ]
        ),
        "sub1": RDFType.Literal(
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
        ),
        "sub2": RDFType.Literal(
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
        ),
    }
    assert sm.mappings.height == 4
    assert sm.mappings.get_column("sub1").to_list() == [
        '""@no',
        '"caAc"@en',
        '""@se',
        None,
    ]
    assert sm.mappings.get_column("sub2").to_list() == [
        '""@no',
        '"c"@en',
        '""@se',
        None,
    ]


@pytest.mark.parametrize("streaming", [True, False])
def test_case_multi_type_with_only_lang(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?ucase ?lcase WHERE {
    VALUES (?a) { ("abcaAc"@en) ("ab"@no) ("bb"@se) (1) }
    BIND(UCASE(?a) as ?ucase)
    BIND(LCASE(?a) as ?lcase)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi(
            [
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
            ]
        ),
        "ucase": RDFType.Literal(
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
        ),
        "lcase": RDFType.Literal(
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
        ),
    }
    assert sm.mappings.height == 4
    assert sm.mappings.get_column("ucase").to_list() == [
        '"AB"@no',
        '"ABCAAC"@en',
        '"BB"@se',
        None,
    ]
    assert sm.mappings.get_column("lcase").to_list() == [
        '"ab"@no',
        '"abcaac"@en',
        '"bb"@se',
        None,
    ]


@pytest.mark.parametrize("streaming", [True, False])
def test_case_multi_type_with_string_and_lang_string_and_other(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?ucase ?lcase WHERE {
    VALUES (?a) { ("abcaAc"@en) ("ab") ("bb"@se) (1) (xsd:abc) }
    BIND(UCASE(?a) as ?ucase)
    BIND(LCASE(?a) as ?lcase)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi(
            [
                RDFType.IRI,
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
        "ucase": RDFType.Multi(
            [
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
        "lcase": RDFType.Multi(
            [
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
    }
    assert sm.mappings.height == 5
    assert sm.mappings.get_column("ucase").to_list() == [
        None,
        '"ABCAAC"@en',
        '"BB"@se',
        None,
        '"AB"',
    ]
    assert sm.mappings.get_column("lcase").to_list() == [
        None,
        '"abcaac"@en',
        '"bb"@se',
        None,
        '"ab"',
    ]


@pytest.mark.parametrize("streaming", [True, False])
def test_before_after_multi_type_with_string_and_lang_string_and_other(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?bef ?aft WHERE {
    VALUES (?a) { ("abcaAc"@en) ("abAcACAc") ("bb"@se) (1) (xsd:abc) }
    BIND(STRBEFORE(?a, "Ac") as ?bef)
    BIND(STRAFTER(?a, "ab") as ?aft)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi(
            [
                RDFType.IRI,
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
        "bef": RDFType.Multi(
            [
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
        "aft": RDFType.Multi(
            [
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
    }
    assert sm.mappings.height == 5
    assert sm.mappings.get_column("bef").to_list() == [
        None,
        '"abca"@en',
        '"bb"@se',
        None,
        '"abAcAC"',
    ]
    assert sm.mappings.get_column("aft").to_list() == [
        None,
        '"caAc"@en',
        '"bb"@se',
        None,
        '"AcACAc"',
    ]


@pytest.mark.parametrize("streaming", [True, False])
def test_before_after_only_lang_string_and_other(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?bef ?aft WHERE {
    VALUES (?a) { ("abcaAc"@en) ("bb"@se) (1) (xsd:abc) }
    BIND(STRBEFORE(?a, "Ac") as ?bef)
    BIND(STRAFTER(?a, "ab") as ?aft)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi(
            [
                RDFType.IRI,
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
            ]
        ),
        "bef": RDFType.Literal("http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"),
        "aft": RDFType.Literal("http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"),
    }
    assert sm.mappings.height == 4
    assert sm.mappings.get_column("bef").to_list() == [
        None,
        '"abca"@en',
        '"bb"@se',
        None,
    ]
    assert sm.mappings.get_column("aft").to_list() == [
        None,
        '"caAc"@en',
        '"bb"@se',
        None,
    ]


@pytest.mark.parametrize("streaming", [True, False])
def test_before_after_only_lang_string_and_other(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?bef ?aft WHERE {
    VALUES (?a) { ("abcaAc"@en) ("bb"@se) }
    BIND(STRBEFORE(?a, "Ac") as ?bef)
    BIND(STRAFTER(?a, "ab") as ?aft)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    print(sm.mappings)
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"),
        "bef": RDFType.Literal("http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"),
        "aft": RDFType.Literal("http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"),
    }
    assert sm.mappings.height == 2
    assert sm.mappings.get_column("bef").to_list() == [
        '"abca"@en',
        '"bb"@se',
    ]
    assert sm.mappings.get_column("aft").to_list() == [
        '"caAc"@en',
        '"bb"@se',
    ]


@pytest.mark.parametrize("streaming", [True, False])
def test_before_after_only_string_and_other(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?bef ?aft WHERE {
    VALUES (?a) { ("abcaAc") ("bb") (1) (xsd:abc) }
    BIND(STRBEFORE(?a, "Ac") as ?bef)
    BIND(STRAFTER(?a, "ab") as ?aft)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi(
            [
                RDFType.IRI,
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
        "bef": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
        "aft": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
    }
    assert sm.mappings.height == 4
    assert sm.mappings.get_column("bef").to_list() == [
        None,
        None,
        "abca",
        "bb",
    ]
    assert sm.mappings.get_column("aft").to_list() == [
        None,
        None,
        "caAc",
        "bb",
    ]


@pytest.mark.parametrize("streaming", [True, False])
def test_substr_multi_type_with_string_and_lang_string_and_other(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?sub1 ?sub2 WHERE {
    VALUES (?a) { ("abcaAc"@en) ("ab") ("bb"@se) (1) (xsd:abc) }
    BIND(SUBSTR(?a, 3,1) as ?sub1)
    BIND(SUBSTR(?a, 1) as ?sub2)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi(
            [
                RDFType.IRI,
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
        "sub1": RDFType.Multi(
            [
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
        "sub2": RDFType.Multi(
            [
                RDFType.Literal(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString"
                ),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
    }
    assert sm.mappings.height == 5
    assert sm.mappings.get_column("sub1").to_list() == [
        None,
        '"a"@en',
        '""@se',
        None,
        '""',
    ]
    assert sm.mappings.get_column("sub2").to_list() == [
        None,
        '"bcaAc"@en',
        '"b"@se',
        None,
        '"b"',
    ]


@pytest.mark.parametrize("streaming", [True, False])
def test_replace_multi(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?replace WHERE {
    VALUES (?a) { ("abcabc") ("ab") (3) }
    BIND(REPLACE(?a, "ab", "ba") as ?replace)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi(
            [
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
        "replace": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
    }
    assert sm.mappings.height == 3
    assert sm.mappings.get_column("replace").to_list() == [None, "ba", "bacbac"]


@pytest.mark.parametrize("streaming", [True, False])
def test_regex_multi(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?replace WHERE {
    VALUES (?a) { ("abcabc") ("ab") (3) }
    BIND(REGEX(?a, "ab") as ?replace)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Multi(
            [
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#integer"),
                RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
            ]
        ),
        "replace": RDFType.Literal("http://www.w3.org/2001/XMLSchema#boolean"),
    }
    assert sm.mappings.height == 3
    assert sm.mappings.get_column("replace").to_list() == [None, True, True]


@pytest.mark.parametrize("streaming", [True, False])
def test_regex(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?replace WHERE {
    VALUES (?a) { ("abcabc") ("ab") ("bb") }
    BIND(REGEX(?a, "ab") as ?replace)
    } ORDER BY ?a
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
        "replace": RDFType.Literal("http://www.w3.org/2001/XMLSchema#boolean"),
    }
    assert sm.mappings.height == 3
    assert sm.mappings.get_column("replace").to_list() == [True, True, False]


@pytest.mark.parametrize("streaming", [True, False])
def test_issue_22(streaming):
    m = Model()
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a (COUNT(*) as ?c) WHERE {
    VALUES (?a) { ("abbb") ("abbb") ("ab") ("a") ("a") }
    BIND(REPLACE(?a, "b", "") as ?replace)
    } GROUP BY ?a 
    ORDER BY ?a 
    """,
        include_datatypes=True,
        streaming=streaming,
    )
    print(sm.mappings)
    assert sm.rdf_types == {
        "a": RDFType.Literal("http://www.w3.org/2001/XMLSchema#string"),
        "c": RDFType.Literal("http://www.w3.org/2001/XMLSchema#unsignedInt"),
    }
    assert sm.mappings.height == 3
    assert sm.mappings.get_column("c").to_list() == [2, 1, 2]



import polars as pl
import pytest

from maplib import (
    Model,
    Template,
    IRI,
    Prefix,
    Triple,
    Variable,
    Parameter,
    xsd,
    RDFType,
    Argument,
)
from polars.testing import assert_frame_equal


def test_want_float_got_int64():
    df = pl.DataFrame({"MyValue": [1]})
    model = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [Parameter(my_value, rdf_type=RDFType.Literal(xsd.float))],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
        ],
    )
    with pytest.raises(Exception):
        model.map(template, df)


def test_want_rdfs_literal_got_int64():
    df = pl.DataFrame({"MyValue": [1]})
    model = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [
            Parameter(
                my_value,
                rdf_type=RDFType.Literal(
                    "http://www.w3.org/2000/01/rdf-schema#Literal"
                ),
            )
        ],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
        ],
    )
    model.map(template, df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.long)


def test_want_rdfs_resource_got_int64():
    df = pl.DataFrame({"MyValue": [1]})
    model = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [
            Parameter(
                my_value,
                rdf_type=RDFType.Literal(
                    "http://www.w3.org/2000/01/rdf-schema#Resource"
                ),
            )
        ],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
        ],
    )
    model.map(template, df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.long)


def test_want_xsd_int_got_xsd_boolean():
    
    df = pl.DataFrame({"MyValue": [True]})
    model = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [Parameter(my_value, rdf_type=RDFType.Literal(xsd.int_))],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
        ],
    )
    with pytest.raises(Exception):
        model.map(template, df)


def test_autoconverted_datetime_to_date():
    
    df = pl.DataFrame({"MyValue": ["2020-02-02T00:00:00Z"]}).cast(pl.Datetime("ns"))
    model = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [Parameter(my_value, rdf_type=RDFType.Literal(xsd.date))],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
        ],
    )
    model.map(template, df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.date)


def test_autoconverted_optional_datetime_to_date():
    
    df = pl.DataFrame({"MyValue": ["2020-02-02T00:00:00Z"]}).cast(pl.Datetime("ns"))
    model = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_other_value = Variable("MyOtherValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [
            Parameter(my_value, rdf_type=RDFType.Literal(xsd.date)),
            Parameter(
                my_other_value, rdf_type=RDFType.Literal(xsd.date), optional=True
            ),
        ],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
        ],
    )
    model.map(template, df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.date)


def test_autoconverted_datetime_list_to_date_list_1():
    
    df = pl.DataFrame({"MyValue": [["2020-02-02T00:00:00Z"]]}).cast(
        pl.List(pl.Datetime("ns"))
    )
    model = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [Parameter(my_value, rdf_type=RDFType.Nested(RDFType.Literal(xsd.date)))],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
        ],
    )
    model.map(template, df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Multi(
        [
            RDFType.IRI,
            RDFType.BlankNode,
            RDFType.Literal("http://www.w3.org/2001/XMLSchema#date"),
        ]
    )


def test_autoconverted_datetime_list_to_date_list_2():
    
    df = pl.DataFrame({"MyValue": [["2020-02-02T00:00:00Z"]]}).cast(
        pl.List(pl.Datetime("ns"))
    )
    model = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [Parameter(my_value, rdf_type=RDFType.Nested(RDFType.Literal(xsd.date)))],
        [
            Triple(
                my_object,
                ex.suf("hasValue"),
                Argument(my_value, list_expand=True),
                list_expander="cross",
            ),
        ],
    )
    model.map(template, df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal("http://www.w3.org/2001/XMLSchema#date")


def test_want_xsd_long_got_xsd_short():
    
    df = pl.DataFrame({"MyValue": [1]})
    df = df.with_columns(pl.col("MyValue").cast(pl.Int16))
    model = Model()
    ex = Prefix("http://example.net/ns#")
    my_value = Variable("MyValue")
    my_object = ex.suf("MyObject")
    template = Template(
        ex.suf("ExampleTemplate"),
        [Parameter(my_value, rdf_type=RDFType.Literal(xsd.long))],
        [
            Triple(my_object, ex.suf("hasValue"), my_value),
        ],
    )
    model.map(template, df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.short)


def test_nested_template_more_general():
    df = pl.DataFrame({"MyValue": ["hello"]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
<http://example.net/ns#ExampleNestedTemplate> [ rdfs:Literal ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ xsd:string ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplate>(?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.string)


def test_nested_template_not_found():
    df = pl.DataFrame({"MyValue": ["hello"]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
<http://example.net/ns#ExampleNestedTemplate> [ rdfs:Literal ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ xsd:string ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplateWrong>(?MyValue)
} . 
    """
    with pytest.raises(Exception):
        m = Model(templates)
        m.map("http://example.net/ns#ExampleTemplate", df)


def test_nested_template_more_specific():
    df = pl.DataFrame({"MyValue": ["hello"]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
<http://example.net/ns#ExampleNestedTemplate> [ xsd:string ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ rdfs:Literal ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplate>(?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.string)


def test_nested_template_more_specific_but_is_IRI():
    df = pl.DataFrame({"MyValue": ["hello"]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
<http://example.net/ns#ExampleNestedTemplate> [ ottr:IRI ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ rdfs:Literal ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplate>(?MyValue)
} . 
    """
    with pytest.raises(Exception):
        m = Model(templates)
        m.map("http://example.net/ns#ExampleTemplate", df)


def test_nested_template_more_general_but_is_IRI():
    df = pl.DataFrame({"MyValue": ["http://example.net/ns#MyValue"]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleNestedTemplate> [ owl:Class ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ xsd:string ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplate>(?MyValue)
} . 
    """
    with pytest.raises(Exception):
        m = Model(templates)
        m.map("http://example.net/ns#ExampleTemplate", df)


def test_nested_template_both_general_but_always_incompatible():
    df = pl.DataFrame({"MyValue": ["http://example.net/ns#MyValue"]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleNestedTemplate> [ owl:Class ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ rdfs:Literal ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplate>(?MyValue)
} . 
    """
    with pytest.raises(Exception):
        m = Model(templates)
        m.map("http://example.net/ns#ExampleTemplate", df)


def test_nested_template_both_are_general_literal_and_compatible():
    df = pl.DataFrame({"MyValue": [1]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleNestedTemplate> [ owl:real ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ rdfs:Literal ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplate>(?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.long)


def test_nested_template_both_are_general_literal_and_compatible_2():
    df = pl.DataFrame({"MyValue": [1]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleNestedTemplate> [ owl:rational ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ owl:real ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplate>(?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.long)


def test_nested_template_both_are_general_literal_and_possibly_but_not_necessarily_incompatible():
    df = pl.DataFrame({"MyValue": [1]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleNestedTemplate> [ owl:real ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ owl:rational ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplate>(?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.long)


def test_nested_template_both_are_general_and_possibly_but_not_necessarily_incompatible():
    df = pl.DataFrame({"MyValue": [1]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleNestedTemplate> [ owl:real ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ rdfs:Resource ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplate>(?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.long)


def test_nested_template_both_are_general_and_compatible_2():
    df = pl.DataFrame({"MyValue": [1]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleNestedTemplate> [ rdfs:Resource ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ owl:real ?MyValue ] :: {
  <http://example.net/ns#ExampleNestedTemplate>(?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.long)


def test_nested_template_both_are_general_and_possibly_but_not_necessarily_incompatible_list():
    df = pl.DataFrame({"MyValue": [[1]]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleNestedTemplate> [ owl:real ?MyValue ] :: {
  <http://ns.ottr.xyz/0.4/Triple>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValue>,?MyValue)
} . 
<http://example.net/ns#ExampleTemplate> [ List<rdfs:Resource> ?MyValue ] :: {
  cross | <http://example.net/ns#ExampleNestedTemplate>(++?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.rdf_types["c"] == RDFType.Literal(xsd.long)


def test_list_arg_to_ottr_triple_should_get_rdf_representation():
    df = pl.DataFrame({"MyValue": [[1]]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleTemplate> [ List<rdfs:Resource> ?MyValue ] :: {
  ottr:Triple(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValueList>,?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.mappings.height == 3
    assert r.rdf_types["c"] == RDFType.Multi(
        [RDFType.IRI, RDFType.BlankNode, RDFType.Literal(xsd.long)]
    )


def test_list_arg_to_ottr_triple_should_get_rdf_representation_nested():
    df = pl.DataFrame({"MyValue": [[1, 2]]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleNestedTemplate> [ ?a, ?b, rdfs:Resource ?c ] :: {
  ottr:Triple(?a, ?b, ?c)
} . 
<http://example.net/ns#ExampleTemplate> [ List<rdfs:Resource> ?MyValue ] :: {
    <http://example.net/ns#ExampleNestedTemplate>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValueList>,?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.mappings.height == 5
    assert r.rdf_types["c"] == RDFType.Multi(
        [RDFType.IRI, RDFType.BlankNode, RDFType.Literal(xsd.long)]
    )


def test_list_arg_to_ottr_triple_should_get_rdf_representation_multiple_executions():
    df1 = pl.DataFrame({"MyValue": [[1, 2]]})
    df2 = pl.DataFrame({"MyValue": [[3, 4]]})
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
<http://example.net/ns#ExampleTemplate> [ List<rdfs:Resource> ?MyValue ] :: {
  ottr:Triple(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValueList>,?MyValue)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate", df1)
    model.map("http://example.net/ns#ExampleTemplate", df2)
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.mappings.height == 10
    assert r.rdf_types["c"] == RDFType.Multi(
        [RDFType.IRI, RDFType.BlankNode, RDFType.Literal(xsd.long)]
    )


def test_constant_arg_has_valid_subtype():
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
<http://example.net/ns#ExampleNestedTemplate> [ ?a, ?b, rdfs:Resource ?c ] :: {
  ottr:Triple(?a, ?b, ?c)
} . 
<http://example.net/ns#ExampleTemplate> [ ] :: {
    <http://example.net/ns#ExampleNestedTemplate>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValueList>, rdf:type)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate")
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.mappings.height == 1
    assert r.rdf_types["c"] == RDFType.IRI


def test_iri_datatype():
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
<http://example.net/ns#ExampleNestedTemplate> [ ?a, ?b, <http://www.w3.org/1999/02/22-rdf-syntax-ns#Resource> ?c ] :: {
  ottr:Triple(?a, ?b, ?c)
} . 
<http://example.net/ns#ExampleTemplate> [ ] :: {
    <http://example.net/ns#ExampleNestedTemplate>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValueList>, rdf:type)
} . 
    """
    model = Model()
    model.add_template(templates)
    model.map("http://example.net/ns#ExampleTemplate")
    r = model.query(
        """
    SELECT ?a ?b ?c WHERE {
        ?a ?b ?c
    }
    """,
        include_datatypes=True,
    )
    assert r.mappings.height == 1
    assert r.rdf_types["c"] == RDFType.IRI


@pytest.mark.skip("Missing general validation logic")
def test_constant_arg_has_invalid_subtype():
    templates = """
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
<http://example.net/ns#ExampleNestedTemplate> [ ?a, ?b, rdfs:Literal ?c ] :: {
  ottr:Triple(?a, ?b, ?c)
} . 
<http://example.net/ns#ExampleTemplate> [ ] :: {
    <http://example.net/ns#ExampleNestedTemplate>(<http://example.net/ns#MyObject>,<http://example.net/ns#hasValueList>, rdf:type)
} . 
    """
    with pytest.raises(Exception):
        Model(templates)
